import re
import os
import sys
import copy
import reflex
import numpy as np
from astropy.io import fits
from astropy.table import Table
import importlib

try :
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except ImportError:
    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except Exception as e:
    raise(e)

# molecfit_wkf generic methods...
# ------------------------------------------------------------------------------------------
def _get_object(header=None) :
    # leave as default from molecfit_model for now (which takes OBS.TARG.NAME -> OBJECT)
    return header.get('OBJECT','UNKNOWN')
# ------------------------------------------------------------------------------------------
def dataset_chooser_keywords() :
    return 'INSTRUME,MJD-OBS,OBJECT,INS.PATH,INS.EXP.MODE,PRO.CATG'
# ------------------------------------------------------------------------------------------
def set_inst_setup(header=None) :
    if header :
        return "%s_%s" %(
            header.get('HIERARCH ESO INS PATH', 'UNKNOWN'),
            header.get('HIERARCH ESO INS EXP MODE', 'UNKNOWN'),
        )
    return 'None'
# ------------------------------------------------------------------------------------------
def check_format(files, output_dir):
    output_files=list()
    files_dict={}
    for file in files:
        if file.category not in files_dict.keys() :
            files_dict[file.category]=[]
        files_dict[file.category]+=[file,]
    has_sci_and_std=(
        ('SCIENCE' in files_dict.keys() or 'SCIENCE_CALCTRANS' in files_dict.keys())
        and
        'STD_MODEL' in files_dict.keys()
    )
    for k in files_dict.keys() :
        for file in files_dict[k] :
            if k in ['STD_MODEL','SCIENCE','SCIENCE_CALCTRANS',] :
                prefix=re.sub('_.*','',k)
                # If this is an IDP format file pass through...
                # If this is a RBNSPECTRA format file then:
                #    a) check for the RBNERRORS file and append data as appropriate...
                #    b) select/prepare a single file for model+calctrans
                #    c) prepare the individual fibres for correct
                hdulist=fits.open(file.name)
                procatg=hdulist[0].header.get('HIERARCH ESO PRO CATG','NULL')
                out_file_name="%s/%s" %(output_dir,os.path.basename(file.name))
                if procatg in [
                    "SCIENCE_RBNSPECTRA",
                    "STD_RBNSPECTRA",
                    ] :
                    #    a) check for the RBNERRORS file and append data as appropriate...
                    hdulist_errs=None
                    if prefix+'_ERRORS' in files_dict.keys() :
                        for f in files_dict[prefix+'_ERRORS'] :
                            ehdulist=fits.open(f.name)
                            if ehdulist[0].header.get('MJD-OBS',-1.) == hdulist[0].header.get('MJD-OBS',0.) :
                                hdulist_errs=ehdulist
                                break

                    # reformat it into a MolecFit Compliant [_MFC] file...
                    model_file, correct_filelist=reformat_RBNSPECTRA(
                        file.name, output_dir, hdulist, hdulist_errs
                    )

                    # broadcast the file to use for molecfit_model...
                    if (
                        ( has_sci_and_std and k == "STD_MODEL" )
                        or
                        ( not has_sci_and_std )
                    ) :
                        output_files.append(reflex.FitsFile(
                            model_file[0], "STD_MODEL", model_file[1], file.purposes
                        ))

                    # broadcast the file to use for molecfit_calctrans...
                    if (
                        ( has_sci_and_std and k != "STD_MODEL" )
                        or
                        ( not has_sci_and_std )
                    ) :
                        output_files.append(reflex.FitsFile(
                            model_file[0], "SCIENCE_CALCTRANS", model_file[1], file.purposes
                        ))

                    # broadcast the original frame (for sticking it all back together at the end...)
                    output_files.append(reflex.FitsFile(
                        file.name, "ORIG_%s" %(file.category), None, file.purposes
                    ))
                elif procatg in [
                    "SCIENCE_RBNSPEC_IDP",
                    ] :
                    # Check for WAVE_FRAME, usually in HELIOCEN for IDPs
                    specsys=hdulist[0].header.get('SPECSYS', 'UNKNOWN')
                    geocorr = hdulist[0].header.get('GEOCORR')
                    helicorr = hdulist[0].header.get('HELICORR')
                    if (
                        specsys in ['HELIOCEN', ]
                        and
                        helicorr is not None
                        and
                        geocorr is not None
                    ) :
                        clam='WAVE'
                        wave_tv=hdulist[1].columns[clam].copy()
                        wave_tv.name='WAVE_TV'
                        wave_tv.array=iu.heli_air_to_topo_vac(
                            hdulist[1].data[clam],
                            geocorr,
                            helicorr,
                            wlg,
                        )
                        new_columns=hdulist[1].columns + fits.ColDefs([wave_tv,])
                        nhdu=fits.BinTableHDU.from_columns(new_columns,)
                        nhdu.header+=hdulist[1].header
                        nhdu.header['HIERARCH ESO MF WAVE COLUMN']='WAVE_TV'
                        nhdu.header['HIERARCH ESO MF WAVE COLUMN WAVELENGTH_FRAME']='TOPO+VAC'
                        hdulist[1]=nhdu
                        iu.HDUList(hdulist).writeto(
                            out_file_name, overwrite=True, checksum=True, output_verify='silentfix',
                        )
                        # broadcast the modified file with WAVE_TV column...
                        output_files.append(reflex.FitsFile(
                            out_file_name, file.category, None, file.purposes
                        ))
                        # broadcast the original frame (for sticking it all back together at the end...)
                        output_files.append(reflex.FitsFile(
                            file.name, "ORIG_%s" %(file.category), None, file.purposes
                        ))
                    else :
                        # broadcast the original file it...
                        output_files.append( reflex.FitsFile(file.name, file.category, None, file.purposes ))
                else :
                    # broadcast the original file it...
                    output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))
            else :
                # broadcast it...
                output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))

    return output_files
# ------------------------------------------------------------------------------------------
wlg = 1.0e-03  # WLG_TO_MICRON
def set_parameters(header=None, hdu=None, previous_inst_setup='None', files=None) :
    wframe='AIR'
    clam='NULL'
    cflux='NULL'
    cdflux='NULL'
    fit_tel_back="False"
    fit_cont="1"
    cont_poly = int(3)
    varkern="FALSE"
    ftol = 1e-10
    xtol = 1e-10

    return {
        'WLG_TO_MICRON': wlg,
        'WAVELENGTH_FRAME': wframe,
        'FTOL': ftol,
        'XTOL': xtol,
        'COLUMN_LAMBDA': clam,
        'COLUMN_FLUX': cflux,
        'COLUMN_DFLUX': cdflux, 
        'FIT_TELESCOPE_BACKGROUND': fit_tel_back,
        'FIT_CONTINUUM': fit_cont,
        'CONTINUUM_N': cont_poly,
        'VARKERN': varkern,
    }
# ------------------------------------------------------------------------------------------
def check_idp(hdu, previous_inst_setup='None', files=None, ParameterInitialization=False, recipe=None, ) :
    no_output=0

    # Data format dependent setting of variables...
    if hdu[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
            'SCIENCE_RBNSPEC_IDP',
            'SCIENCE_RBNSPEC_STACK',
            'SCIENCE_RBNSPEC_MFC',
            'STD_RBNSPEC_MFC',
            'SCIENCE_RBNSPECTRA_MFC',
            'STD_RBNSPECTRA_MFC',
        ] :
        # IDP=binary-table-format format...
        # Obtainable from the ESO Science Archive facility
        # or by reducing with "--generate-SDP-format=true" with ESO PL
        # (Set WKF variable "" = true)
        # MFC = MolecFit Compatible, generated from ESO PL 2-d products
        # by 
        clam="WAVE"
        wframe='AIR'
        if "WAVE_TV" in hdu[1].columns.names :
            clam="WAVE_TV"
            wframe='VAC'

        cflux="FLUX_REDUCED"
        cdflux="ERR_REDUCED"
        primary='FALSE'
        errorext=0
        maskext=0
        map_convolve="1,1"
        map_atmosphere="0,1"
        map_correct="0,1"
        dd=hdu[1].data[cflux]
        wl_range=np.array([np.min(hdu[1].data[clam][0]),np.max(hdu[1].data[clam][0])])
        wl_resolution=np.max(hdu[1].data[clam][0][1:]-hdu[1].data[clam][0][0:-1])
    else:
        # these are the values to use for FLAMES spectra in generic PL format (no IDP)
        ## This section not fully implemented yet, so far we assume use of IDP versions
        raise Exception('Non-IDP-format=non-binary-table-format input files not (yet) supported.')
        primary='TRUE'
        errorext=0
        maskext=0
        map_convolve="1"
        map_atmosphere="1"
        map_correct="1"
        median_cont=3000.
        try:
            dd=hdu[0].data
        except:         
            dd=hdu[1].data
        wl_range=np.array([hdu[0].header['CRVAL1'],hdu[0].header['CRVAL1']+hdu[0].header['NAXIS1']*hdu[0].header.get('CD1_1',hdu[0].header.get('CDELT1'))])
        wl_resolution=hdu[0].header.get('CD1_1',hdu[0].header.get('CDELT1'))
    median_cont = float(np.median(dd[np.where(~np.isnan(dd))]))

    WL_inc_exc_PIX_exc={}
    inst_setup_changed = set_inst_setup(hdu[0].header) != previous_inst_setup
    if ParameterInitialization :
        ## ---------------------------------------------------------------------
        # Automatically set molecules and ranges...
        wl_span=wl_range[1]-wl_range[0]

        fallback_WL_include_ranges={
                ## Following is "selected by hand" but from XSHOOTER VIS arm only...
                ## i.e. good over WL-range ~[550.0,1000.0].
                'H2O': [
                    [585.,602.],
                    [645.,662.],
                    [690.,743.],
                    [785.,846.],
                    [880.,997.],
                ],
                'CO2': [
                    [780.,788.],
                    [788.,795.],
                    [868.,878.],
                ],
                'O3': [
                    [535.,800.],
                ],
                'N2O': [
                    #[,],
                ],
                'CO': [
                    #[,],
                ],
                'CH4': [
                    [880.5,901.5],
                ],
                'O2': [
                    [555.,600.],
                    [625.,635.],
                    [685.,698.],
                    [756.,778.],
                ],
            }

        WL_ranges_to_include=am.auto_molecules_set_WL_ranges_to_include(
            files,
            wl_range,
            wl_span,
            wl_resolution,
            wl_scale=1./wlg,
            buff=wl_span*0.05,
            fallback_WL_include_ranges=fallback_WL_include_ranges,
        )
        WL_inc_exc_PIX_exc={
            'LIST_MOLEC': WL_ranges_to_include['list_mol'],
            'FIT_MOLEC': WL_ranges_to_include['fit_mol'],
            'REL_COL': WL_ranges_to_include['rel_col'],
            'WAVE_INCLUDE': WL_ranges_to_include['wave_include'],
            'WAVE_EXCLUDE': WL_ranges_to_include['wave_exclude'],
            'PIXEL_EXCLUDE': WL_ranges_to_include['pixel_exclude'],
        }
        ## ---------------------------------------------------------------------

    return dict(
        list(({
            # model settings...
            'WAVELENGTH_FRAME': wframe,
            'COLUMN_LAMBDA': clam,
            'COLUMN_FLUX': cflux,
            'COLUMN_DFLUX': cdflux,
            'USE_ONLY_INPUT_PRIMARY_DATA': primary,
            'USE_DATA_EXTENSION_AS_DFLUX': errorext,
            'USE_DATA_EXTENSION_AS_MASK': maskext,
            'CONTINUUM_CONST': median_cont,
            # calctrans settings...
            'MAPPING_CONVOLVE': map_convolve,
            'MAPPING_ATMOSPHERIC': map_atmosphere,
            # correct settings...
            'MAPPING_CORRECT': map_correct,
            # Only set the following if the canvas value is NULL
            'ignore_if_not_NULL': {} if inst_setup_changed else WL_inc_exc_PIX_exc,
        }).items())
        +
        list((WL_inc_exc_PIX_exc if inst_setup_changed else {}).items())
    )
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
# -----   GIRAFFE specific methods   -------------------------------------------------------
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
def reformat_RBNSPECTRA(filename, output_dir, hdulist, hdulist_errs=None) :
    '''
    Convert a single, multi-fibre image format SCIENCE|STD_RBNSTECTRA input file
    (plus optional SCIENCE|STD_RBNERRORS into)
    bin-table format files, one per fibre/
    '''
    # Initialise list of returned files...
    correct_filelist=[]

    # Compute wave array from header keywords...
    wave=(
        (np.arange(1, hdulist[0].header['NAXIS2']+1, 1)-hdulist[0].header['CRPIX2'])
        *
        hdulist[0].header['CDELT2'] + hdulist[0].header['CRVAL2']
    )
    fibre_setup_tbl=Table.read(filename, format='fits')
    snr=[]
    for _f in range(hdulist[0].header['NAXIS1']) :
        snr+=[iu.DER_SNR(hdulist[0].data[:,_f]),]
    # Select/prepare the fibre/combination of fibres to use for
    #  - molecfit_model
    #  - molecfot_calctrans
    #
    ## For the moment we only have one method: eventually the strategy
    ## could/should be user selectable in the main canvas...
    ## e.g.
    ##   1. fibre with the highest S/N
    ##   2. 'pseudo-fibre' from combination of fibres as long as S/N increases,
    ##      addign fibres by decresaing S/N...
    ##   3. 'pseudo-fibre' from combination of SKY fibres...
    model_file_selection_method="fib_with_max_SNR"
    ##
    if model_file_selection_method == "fib_with_max_SNR" :
        ## Simple first case: the single fibre with the highest S/N...
        max_snr_index=np.argsort(snr)[-1]
        flux=hdulist[0].data[:,max_snr_index]
        skyback=hdulist['SKYBACK'].data[:,max_snr_index] if 'SKYBACK' in hdulist else None
        flux_sky_sub=hdulist['FLUX_SKY_SUB'].data[:,max_snr_index] if 'FLUX_SKY_SUB' in hdulist else None
        eflux=flux*0.+1.
        eskyback=flux*0.+1.
        eflux_sky_sub=flux*0.+1.
        if hdulist_errs is not None :
            eflux=hdulist_errs[0].data[:,max_snr_index]
            eskyback=hdulist_errs['SKYBACK_ERR'].data[:,max_snr_index] if 'SKYBACK_ERR' in hdulist_errs else None
            eflux_sky_sub=hdulist_errs['FLUX_SKY_SUB_ERR'].data[:,max_snr_index] if 'FLUX_SKY_SUB_ERR' in hdulist_errs else None

        qual=flux*0
        new_filename="%s/%s_FPS=%03d.fits" %(
            output_dir,
            os.path.basename(re.sub('\.fits$','',filename)),
            fibre_setup_tbl['FPS'][max_snr_index],
        )

    hdr=copy.deepcopy(hdulist[0].header)
    hdr['PRODCATG']='SCIENCE.SPECTRUM'
    orig_catg=hdr.get('HIERARCH ESO PRO CATG','SCIENCE_RBNSPECTRA')
    # _MFC : MolecFit Compatible... similar to and IDP, but not exactly...
    hdr['HIERARCH ESO PRO CATG']='%s_MFC' %(orig_catg)
    fmt=str(len(wave))+'D'
    c1=fits.Column(name='WAVE', array=[wave], format=fmt, unit=hdulist[0].header['CUNIT2'])
    c2=fits.Column(name='FLUX_REDUCED', array=[flux], format=fmt, unit=hdulist[0].header['BUNIT'])
    c3=fits.Column(name='ERR_REDUCED', array=[eflux], format=fmt, unit=hdulist[0].header['BUNIT'])
    cols=[c1,c2,c3]
    # wave is in air for RBNSPECTRA
    # Transform to topo_vac for molecfit_model
    wave_tv=iu.air_to_vac(wave,wlg)
    c4=fits.Column(name='WAVE_TV', array=[wave_tv], format=fmt, unit=hdulist[0].header['CUNIT2'])
    cols+=[c4,]

    if skyback is not None :
        c5=fits.Column(name='SKYBACK', array=[skyback], format=fmt, unit=hdulist[0].header['BUNIT'])
        c6=fits.Column(name='SKYBACK_ERR', array=[eskyback], format=fmt, unit=hdulist[0].header['BUNIT'])
        cols+=[c5,c6]
    if flux_sky_sub is not None :
        c7=fits.Column(name='FLUX_SKY_SUB', array=[flux_sky_sub], format=fmt, unit=hdulist[0].header['BUNIT'])
        c8=fits.Column(name='FLUX_SKY_SUB_ERR', array=[eflux_sky_sub], format=fmt, unit=hdulist[0].header['BUNIT'])
        cols+=[c7,c8]
    
    h0=fits.PrimaryHDU(data=None,header=hdr)

    h1 = fits.BinTableHDU.from_columns(cols)
    h1.header['EXTNAME']='SPECTRUM'
    '''
    h1.header['FPS']=fibre_setup_tbl['FPS'][_f]
    h1.header['SSN']=fibre_setup_tbl['SSN'][_f]
    h1.header['SNR']=snr[_f]
    h1.header['COLUMNID']=_f
    h1.header['ORIGCATG']=orig_catg
    if (
        ( hdr['HIERARCH ESO INS MODE'].strip() not in ["ARG",] )
        or
        ( 'Sky' in fibre_setup_tbl['Retractor'][_f] )
    ):
        h1.header['RA']=fibre_setup_tbl['RA'][_f]
        h1.header['DEC']=fibre_setup_tbl['DEC'][_f]
        h1.header['OBJECT']=fibre_setup_tbl['OBJECT'][_f]
    '''
    h1.header['NELEM']=hdulist[0].header['NAXIS2']
    h1.header['HIERARCH ESO MF WAVE COLUMN']='WAVE_TV'
    h1.header['HIERARCH ESO MF WAVE COLUMN WAVELENGTH_FRAME']='TOPO+VAC'
    '''
    # Create a "file name" keyword that will always fit and allow matching of the
    # input files to products...
    h1.header['HIERARCH ESO DRS MF PRO FILE']="%s.%03d" %(hdulist[0].header['CHECKSUM'],fibre_setup_tbl['FPS'][_f])
    h0.header['HIERARCH ESO DRS MF PRO FILE']="%s.%03d" %(hdulist[0].header['CHECKSUM'],fibre_setup_tbl['FPS'][_f])
    '''
    hl=fits.HDUList([h0,h1])
    print('writing %s' %(new_filename))
    iu.HDUList(hl).writeto(new_filename, overwrite=True, checksum=True, output_verify='silentfix')
    correct_filelist+=[(new_filename,hl[0].header['CHECKSUM']),]

    model_file=correct_filelist[0]

    return model_file, correct_filelist
# ------------------------------------------------------------------------------------------
def recombine_idps( files=None, orig_files=None ) :
    o_files=[]
    cube_files=[]
    TELLURIC_CORR_file=None
    for file in orig_files or {} :
        if "ERRORS" in file.category :
            #print(file)
            pass
        if "ORIG_" in file.category[0:5] :
            hdulist=fits.open(file.name)
            procatg=hdulist[0].header.get('HIERARCH ESO PRO CATG','NULL')
            if procatg in [
                "SCIENCE_RBNSPECTRA",
                "STD_RBNSPECTRA",
                ] :
                o_files+=[file,]
        elif file.category in [
            "SCIENCE_CUBE_SPECTRA",
            "STD_CUBE_SPECTRA",
            ] :
            cube_files+=[file]
        elif file.category in [
            "TELLURIC_CORR",
            ] :
            TELLURIC_CORR_file=file
    for file in o_files :
        return recombine_idps_RBNSPECTRA(
            file,
            hdulist,
            files,
            cube_files,
            TELLURIC_CORR_file,
        )
    return files
# ------------------------------------------------------------------------------------------
def recombine_idps_RBNSPECTRA(
        orig_file,
        hdulist,
        files=None,
        cube_files=None,
        TELLURIC_CORR_file=None,
    ) :
    new_files=None
    propagate_keys_file=None
    TELLURIC_CORR={
        "SCIENCE_TELLURIC_CORR": fits.open(hdulist.filename()),
        # Don't bother with the SPECTRUM_TELLURIC_CORR, it seems to be an exact copy
        # of SCIENCE_TELLURIC_CORR...
        #"SPECTRUM_TELLURIC_CORR": copy.deepcopy(hdulist),
    }
    if "SCIENCE_TELLURIC_CORR" in TELLURIC_CORR.keys() :
        #TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][0].data*=0.

        # Put the SCIENCE_TELLURIC_CORR into the PrimaryHDU
        # and the SPECTRUM_TELLURIC_CORR into an added HDU on the end...
        # For convenience, copy the original RBNSPECTRA into another HDU
        TELLURIC_CORR["SCIENCE_TELLURIC_CORR"].insert(2, fits.ImageHDU(data=copy.deepcopy(hdulist[0].data), header=hdulist[0].header))
        TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][2].header['EXTNAME']=TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][2].header['ESO PRO CATG']
        del TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][2].header['ESO PRO CATG']
        del TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][2].header['PIPEFILE']
        if "SPECTRUM_TELLURIC_CORR" in TELLURIC_CORR.keys() :
            # And prepare another HDU for the SPECTRUM_TELLURIC_CORR...
            TELLURIC_CORR["SCIENCE_TELLURIC_CORR"].append(fits.ImageHDU(data=copy.deepcopy(hdulist[0].data), header=hdulist[0].header))
            TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][-1].header['EXTNAME']="SPECTRUM_TELLURIC_CORR"
            del TELLURIC_CORR["SCIENCE_TELLURIC_CORR"]["SPECTRUM_TELLURIC_CORR"].header['ESO PRO CATG']
            del TELLURIC_CORR["SCIENCE_TELLURIC_CORR"]["SPECTRUM_TELLURIC_CORR"].header['PIPEFILE']
            TELLURIC_CORR["SCIENCE_TELLURIC_CORR"]["SPECTRUM_TELLURIC_CORR"].data*=0.
        # set PRO.CATG
        TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][0].header['HIERARCH ESO PRO CATG']="SCIENCE_TELLURIC_CORR"

    if "SPECTRUM_TELLURIC_CORR" in TELLURIC_CORR.keys() :
        TELLURIC_CORR["SPECTRUM_TELLURIC_CORR"][0].header['HIERARCH ESO PRO CATG']="SPECTRUM_TELLURIC_CORR"
        TELLURIC_CORR["SPECTRUM_TELLURIC_CORR"][0].data*=0.
    
    # Open the atm transmission file for later use...
    if TELLURIC_CORR_file is not None :
        # Actually not possible not to have TELLURIC_CORR_file I think...
        at_hdulist=fits.open(TELLURIC_CORR_file.name)
        # Don't use /= in order to get double precision...
        TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][0].data=(
            copy.deepcopy(TELLURIC_CORR["SCIENCE_TELLURIC_CORR"][0].data)
            /
            np.transpose(np.array([at_hdulist[1].data]))
        )
        for file in files or {} :
            if file.category in ["SCIENCE_TELLURIC_CORR",] :
                propagate_keys_file=propagate_keys_file or file
    else :
        ## JPr: 2021-06-10
        ## I've turned off using molecfit_correct to correct all the spectra,
        ## instead it will just correct the one (SCIENCE_CALCTRANS)
        ## We should NEVER arrive here...
        raise Exception("Could not find TELLURIC_CORR_file in input dataset.")
        """
        for file in files or {} :
            if file.category in TELLURIC_CORR.keys() :
                tc_hdulist=fits.open(file.name)
                print("Inserting %s..." %file.name)
                if isinstance(tc_hdulist[1],fits.BinTableHDU) :
                    TELLURIC_CORR[file.category][0].data[:,tc_hdulist[1].header['COLUMNID']]=tc_hdulist[1].data['FLUX_REDUCED'][0]
                    propagate_keys_file=propagate_keys_file or file
                else :
                    if TELLURIC_CORR[file.category][0].data is None :
                        TELLURIC_CORR[file.category][0].data=np.zeros(
                            [
                                tc_hdulist[1].header["NAXIS1"],
                                hdulist[0].header["NAXIS1"],
                            ]
                        )
                    # As of ver 4.1.0b2 the number of pixels in the SCIENCE_TELLURIC_CORR gets set to
                    # the number of bytes required to store the bin-table, i.e MUCH too large,
                    # so we "fix" the files here resizing the data array to just the right size.
                    if tc_hdulist[1].header['NAXIS1'] > TELLURIC_CORR['SCIENCE_TELLURIC_CORR'][-1].header['NAXIS2'] :
                        print("resizing %s" %(file.name))
                        tc_hdulist[1].data=tc_hdulist[1].data[0:TELLURIC_CORR['SCIENCE_TELLURIC_CORR'][-1].header['NAXIS2']]
                        iu.HDUList(tc_hdulist).writeto(file.name, overwrite=True, output_verify='silentfix', checksum=True)
                    TELLURIC_CORR[file.category][0].data[:,tc_hdulist[1].header['COLUMNID']]=tc_hdulist[1].data[0:]
                    TELLURIC_CORR['SCIENCE_TELLURIC_CORR']["SPECTRUM_TELLURIC_CORR"].data[:,tc_hdulist[1].header['COLUMNID']]=tc_hdulist[1].data[0:]
        """

    output_dir=os.path.dirname(propagate_keys_file.name or files[0].name)
    bname=os.path.basename(orig_file.name)
    hdr=None
    if propagate_keys_file is not None :
        hdr=iu.propagate_keys(
            propagate_keys_file,
            ["molecfit_model","molecfit_calctrans","molecfit_correct"],
            orig_file
        )
    for k in TELLURIC_CORR.keys() :
        new_files=new_files or []
        out_file_name="%s/%s_%s" %(output_dir,k,bname)
        if hdr is not None :
            TELLURIC_CORR[k][0].header+=hdr
        print("writing '%s'..." %out_file_name)
        TELLURIC_CORR[k][0].header['PIPEFILE']="%s_%s" %(k,bname)
        iu.HDUList(TELLURIC_CORR[k]).writeto(
            out_file_name, overwrite=True, checksum=True, output_verify='silentfix',
        )
        print("wrote '%s'." %out_file_name)
        new_files+=[
            reflex.FitsFile(
                out_file_name,
                k,
                TELLURIC_CORR[k][0].header['CHECKSUM'],
                ['UNIVERSAL',]
            )
        ]

    if TELLURIC_CORR_file is not None :
        for cfile in cube_files or [] :
            new_files=new_files or []
            k="SCIENCE_TELLURIC_CORR"
            bname=os.path.basename(cfile.name)
            c_hdulist=fits.open(cfile.name)
            c_hdulist[0].header+=hdr
            c_hdulist[0].header['ESO PRO CATG']="%s_CUBE" %k
            # The image cube of an ARGUS mode observation is in the Primary HDU
            # The <N> image cubes of an IFU mode observation are in <N> ImageHDUs
            # with the Primary HDU "empty".
            # Both type also have an attached FIBER_SETUP binary table ()
            # so we can't just process all HDUs...
            ## ARGUS
            list_of_c_hdus=[c_hdulist[0]]
            ## IFU
            if c_hdulist[0].header.get('ESO INS MODE','').strip() in ["IFU",] :
                list_of_c_hdus=c_hdulist[1:]

            for c_hdu in list_of_c_hdus :
                if isinstance(c_hdu, (fits.ImageHDU,fits.PrimaryHDU)) :
                    # Don't use /= in order to get double precision...
                    c_hdu.data=c_hdu.data/np.transpose(np.array([[at_hdulist[1].data]]))

            out_file_name="%s/%s_%s" %(output_dir,k,bname)
            print("writing '%s'..." %out_file_name)
            iu.HDUList(c_hdulist).writeto(
                out_file_name,
                overwrite=True,
                checksum=True,
                output_verify='silentfix',
            )
            print("Wrote '%s'..." %out_file_name)
            new_files+=[reflex.FitsFile(out_file_name, k, c_hdulist[0].header['CHECKSUM'], ['UNIVERSAL',])]
    return new_files or files
# ------------------------------------------------------------------------------------------
