# ------------------------------------------------------------------------------------------
import re
import os
import sys
import numpy as np
from astropy.io import fits
import reflex
import importlib

try :
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except ImportError:
    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except Exception as e:
    raise(e)

# ------------------------------------------------------------------------------------------
def set_inst_setup(header=None) :
    if header :
        return "GRAT=%s_WL=%3.1f" %(
            header.get('HIERARCH ESO INS GRAT1 NAME', 'UNKNOWN'),
            header.get('HIERARCH ESO INS GRAT1 WLEN', -1.0),
        )
    return 'None'
# ------------------------------------------------------------------------------------------
def check_format(files,output_dir):
    output_files=list()
    for file in files:
        if file.category in ['SCIENCE','SCIENCE_CALCTRANS','STD_MODEL',]:
            hdu=fits.open(file.name)
            prodcatg=hdu[0].header.get('PRODCATG','NULL')
            out_file_name=file.name
            if prodcatg not in [
                "SCIENCE.SPECTRUM",
            ] :
                # If not IDP, I need to do something: I call instrument-specific modules
                # This should actually be ALL VISIR images, since (at the moment) VISIR doesn't
                # have IDPs.
                out_file_name = output_dir+'/'+(file.name[:-5]+'_edited.fits').rsplit('/',1)[1]
                os.system('cp '+file.name+' '+out_file_name)
                check_format_of_file(out_file_name)

            # broadcast it...
            output_files.append(reflex.FitsFile(out_file_name, file.category, None, file.purposes))
    return output_files
# ------------------------------------------------------------------------------------------
def check_format_of_file(filename) :

    # This module is called only for files that are not in IDP format (single raw)
    # The input file (a copy of the original in the input raw directory) is overwritten
    hdu=fits.open(filename,method='update')
    
    _c_units={}
    for _c in hdu[1].columns.columns :
        if hasattr(_c,'unit') :
            _c_units[_c.name]=_c.unit
    wave=hdu[1].data['WLEN']#getw(hdu[0].header,axis=1)
    flux=hdu[1].data['SPC_EXTRACTED']
    try:
        error=hdu[1].data['SPC_ERROR']
    except:
        error=flux*0+1

    qual=flux*0
    try:
        spc_sky= hdu[1].data['SPC_SKY'] 
        spc_model_xc= hdu[1].data['SPC_MODEL_XC']
        spc_model_ph= hdu[1].data['SPC_MODEL_PH']
        spc_calibrated=hdu[1].data['SPC_MODEL_CALIBRATED']
        spc_calibrated_error=hdu[1].data['SPC_MODEL_CALIBRATED_ERROR']
    except:
        spc_sky=  flux*0+1
        spc_model_xc=  flux*0+1
        spc_model_ph=  flux*0+1
        spc_calibrated=flux*0+1
        spc_calibrated_error=flux*0+1
    hdr=hdu[0].header
    hdr['PRODCATG']='SCIENCE.SPECTRUM'
    
    #hdu[0].header['PRO.CATG']='SCIENCE.SPECTRUM'
    fmt=str(len(wave))+'D'
    oc_names={
        'WAVE': 'WLEN',
        'FLUX': 'SPC_EXTRACTED',
        'ERR': 'SPC_ERROR',
    }
    c1=fits.Column(name='WAVE', array=[wave], format=fmt)
    c2=fits.Column(name='FLUX', array=[flux], format=fmt)
    c3=fits.Column(name='ERR', array=[error], format=fmt)
    c4=fits.Column(name='QUAL', array=[qual], format=fmt)
    c5=fits.Column(name='SPC_MODEL_XC', array=[spc_model_xc], format=fmt)
    c6=fits.Column(name='SPC_MODEL_PH', array=[spc_model_ph], format=fmt)
    c7=fits.Column(name='SPC_MODEL_CALIBRATED', array=[spc_calibrated], format=fmt)
    c8=fits.Column(name='SPC_MODEL_CALIBRATED_ERROR', array=[spc_calibrated_error], format=fmt)
    
    h0=fits.PrimaryHDU(data=None,header=hdr)
    hdr['EXTNAME']='SPECTRUM'
    t = fits.BinTableHDU.from_columns([c1, c2, c3, c4, c5, c6, c7, c8],header=hdr)
    for c in t.columns.columns :
        oc_name=oc_names.get(c.name,c.name)
        if oc_name in _c_units :
            c.unit=_c_units[oc_name]
    hl=fits.HDUList([h0,t])
    iu.HDUList(hl).writeto(filename, overwrite=True, checksum=True, output_verify='silentfix')
# ------------------------------------------------------------------------------------------
wlg = 1000000.
def set_parameters(header=None, hdu=None, previous_inst_setup='None', files=None) :
    #this are the values to use for VISIR spectra, no matter if IDP or not
    wframe='VAC'
    list_mol='H2O,O3'
    fit_mol='1,1'
    rel_col='1.0,1.0'
    clam='WAVE'
    cflux='FLUX'
    cdflux='ERR'
    fit_tel_back="False"
    cont_poly = int(1)
    fit_wlc = "1"
    fit_cont="1"
    wlc_n=int(2)
    ftol=0.01
    xtol=0.01

    slit_width_keyword="ESO INS SLIT1 WID"
    #- Exclude the first and last 15 pixels...
    # we rely on the fact that ALL VISIR data are
    # transformed into Binary tables in check_format_of_file()
    data=hdu[1].data
    lam=data['WAVE'][0]
    wmn=lam[5]*wlg
    wmx=lam[-5]*wlg
    PIXEL_EXCLUDE="0,15,"+str(int(len(lam)-15))+","+str(int(len(lam)-1))
    return {
        'WLG_TO_MICRON': wlg,
        'WAVELENGTH_FRAME': wframe,
        'LIST_MOLEC': list_mol,
        'FIT_MOLEC': fit_mol,
        'REL_COL': rel_col,
        'COLUMN_LAMBDA': clam,
        'COLUMN_FLUX': cflux,
        'COLUMN_DFLUX': cdflux,
        'FIT_TELESCOPE_BACKGROUND': fit_tel_back,
        'CONTINUUM_N': cont_poly,
        'FIT_WLC': fit_wlc,
        'WLC_N': wlc_n,
        'FIT_CONTINUUM': fit_cont,
        #'PIXEL_EXCLUDE':PIXEL_EXCLUDE, # PIPE-9556
        'PIXEL_EXCLUDE': 'NULL',
        'FTOL': ftol,
        'XTOL': xtol,
        'SLIT_WIDTH_KEYWORD': slit_width_keyword,
    }
# ------------------------------------------------------------------------------------------
def check_idp(hdu, previous_inst_setup='None', files=None, ParameterInitialization=False, recipe=None, ) :
    import numpy as np
    no_output=0 
    # these are the values to use for VISIR spectra
    # VISIR does not curently have IDP products, and the PL products aren't
    # compatible with molecfit as-is, so...
    # the input file format has been changed into IDP-like (binary table with 1 vector raw)
    clam='WAVE'
    cflux='FLUX'
    cdflux='ERR'
    primary='FALSE'
    errorext=0
    maskext=0
    map_convolve="1,1"
    map_atmosphere="0,1"
    map_correct="0,1"
    d=hdu[1].data
    median_cont = float(1.)

    #- GENERATE 4 WAVELENGTHS CHUNCKS EQUALLY SPACED
    # we rely on the fact that ALL VISIR data are
    # transformed into Binary tables in check_format_of_file()
    data=hdu[1].data
    lam=data['WAVE'][0]
    wmn=lam[5]*wlg
    wmx=lam[-5]*wlg
    #gap=0.0001 # with this value the gap is smaller than a wl_step which makes it difficult to see where the regions are...
    gap=((wmx-wmn)/(len(lam)-10)*2.)
    bins=5
    a=np.arange(wmn,wmx,(wmx-wmn)/bins)
    b=np.arange(wmn+gap,wmx+gap,(wmx-wmn)/bins)
    b=b[1:]
    c=np.concatenate((a,b,np.array([wmx])))
    c=np.sort(c)
    cstr=str(c).strip()
    cstr=cstr.replace("["," ")
    cstr=cstr.replace("]"," ")
    cstr=" ".join(cstr.split())
    cstr=cstr.strip()
    #cstr='"'+cstr+'"'
    cstr=cstr.replace(" ",",")
    WAVE_INCLUDE=cstr
    PIXEL_EXCLUDE="0,15,"+str(int(len(lam)-15))+","+str(int(len(lam)-1))

    WL_inc_exc_PIX_exc={}
    inst_setup_changed = set_inst_setup(hdu[0].header) != previous_inst_setup
    if ParameterInitialization :
        ## ---------------------------------------------------------------------
        # Automatically set molecules and ranges...
        wl_range=np.array([np.min(hdu[1].data[clam][0]),np.max(hdu[1].data[clam][0])])
        wl_resolution=np.min(hdu[1].data[clam][0][1:]-hdu[1].data[clam][0][0:-1])
        wl_span=wl_range[1]-wl_range[0]

        WL_ranges_to_include=am.auto_molecules_set_WL_ranges_to_include(
            files,
            wl_range,
            wl_span,
            wl_resolution/10., # scale wl_resolution because VISIR is low-res...
            wl_scale=1./wlg,
            buff=wl_span*0.05,
        )
        WL_inc_exc_PIX_exc={
            'LIST_MOLEC': WL_ranges_to_include['list_mol'],
            'FIT_MOLEC': WL_ranges_to_include['fit_mol'],
            'REL_COL': WL_ranges_to_include['rel_col'],
            'WAVE_INCLUDE':WAVE_INCLUDE,
            #'PIXEL_EXCLUDE':PIXEL_EXCLUDE, # PIPE-9556
            'WAVE_EXCLUDE': 'NULL',
            'PIXEL_EXCLUDE': 'NULL',
            #'WAVE_INCLUDE': WL_ranges_to_include['wave_include'],
            #'WAVE_EXCLUDE': WL_ranges_to_include['wave_exclude'],
            #'PIXEL_EXCLUDE': WL_ranges_to_include['pixel_exclude'],
        }
        ## ---------------------------------------------------------------------

    return dict(
        list(({
            # model settings...
            'COLUMN_LAMBDA': clam,
            'COLUMN_FLUX': cflux,
            'COLUMN_DFLUX': cdflux,
            'USE_ONLY_INPUT_PRIMARY_DATA': primary,
            'USE_DATA_EXTENSION_AS_DFLUX': errorext,
            'USE_DATA_EXTENSION_AS_MASK': maskext,
            'CONTINUUM_CONST': median_cont,
            # calctrans settings...
            'MAPPING_CONVOLVE': map_convolve,
            'MAPPING_ATMOSPHERIC': map_atmosphere,
            # correct settings...
            'MAPPING_CORRECT': map_correct,
            # Only set the following if the canvas value is NULL
            'ignore_if_not_NULL': {} if inst_setup_changed else WL_inc_exc_PIX_exc,
        }).items())
        +
        list((WL_inc_exc_PIX_exc if inst_setup_changed else {}).items())
    )
# ------------------------------------------------------------------------------------------
